(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)

Require Import Language_spec. 
Require Import Table.

(* ------------------------   SMALL-STEP SEMANTICS  ------------------------*)

Inductive is_value : expr -> Prop :=
  | Cst_val : forall c : Constante, is_value (Const c)
  | Loc_val : forall l : location, is_value (Loc l)
  | Fun_val : forall (i : identifier) (e : expr), is_value (Fun i e).

Inductive value : Set :=
    value_intro : forall e : expr, is_value e -> value.

Definition value2expr (v : value) := match v with
                                     | value_intro e V => e
                                     end.


Definition context := expr -> expr.

Inductive is_context : context -> Prop :=
  | hole : is_context (fun x : expr => x)
  | app_left :
      forall (e : expr) (c : context),
      is_context c -> is_context (fun x : expr => Apply (c x) e)
  | app_right :
      forall (v : expr) (c : context),
      is_context c ->
      is_value v -> is_context (fun x : expr => Apply v (c x))
  | let_left :
      forall (e : expr) (y : identifier) (c : context),
      is_context c -> is_context (fun x : expr => Let_in y (c x) e)
  | sequ_left :
      forall (e : expr) (c : context),
      is_context c -> is_context (fun x : expr => Sequ (c x) e)
  | ref_ctx :
      forall c : context,
      is_context c -> is_context (fun x : expr => Ref (c x))
  | deref_ctx :
      forall c : context,
      is_context c -> is_context (fun x : expr => Deref (c x))
  | assign_right :
      forall (l : location) (c : context),
      is_context c -> is_context (fun x : expr => Assign (Loc l) (c x))
  | assign_left :
      forall (e : expr) (c : context),
      is_context c -> is_context (fun x : expr => Assign (c x) e)
  | if_left :
      forall (e1 e2 : expr) (c : context),
      is_context c -> is_context (fun x : expr => If (c x) e1 e2).


Definition memory := table location value.
Definition configuration := (expr * memory)%type.

Inductive red_epsilon : configuration -> configuration -> Prop :=
  | beta1 :
      forall (x : identifier) (e er v : expr) (m : memory),
      is_value v ->
      subst_expr e x v er -> red_epsilon (Apply (Fun x e) v, m) (er, m)
  | beta2 :
      forall (x : identifier) (e er v : expr) (m : memory),
      is_value v ->
      subst_expr e x v er -> red_epsilon (Let_in x v e, m) (er, m)
  | red_ref :
      forall (m : memory) (v : expr) (l : location) (V : is_value v),
      ~ in_dom l m ->
      red_epsilon (Ref v, m) (Loc l, add_table l (value_intro v V) m)
  | red_deref :
      forall (m : memory) (l : location) (ml : value),
      in_dom l m ->
      apply_table m l ml -> red_epsilon (Deref (Loc l), m) (value2expr ml, m)
  | red_assign :
      forall (m : memory) (v : expr) (l : location) (V : is_value v),
      in_dom l m ->
      red_epsilon (Assign (Loc l) v, m)
        (Const unit, add_table l (value_intro v V) m)
  | red_sequ :
      forall (m : memory) (e v : expr),
      is_value v -> red_epsilon (Sequ v e, m) (e, m)
  | red_if_true :
      forall (m : memory) (e1 e2 : expr),
      red_epsilon (If vrai e1 e2, m) (e1, m)
  | red_if_false :
      forall (m : memory) (e1 e2 : expr),
      red_epsilon (If faux e1 e2, m) (e2, m)
  | red_while :
      forall (m : memory) (e1 e2 : expr),
      red_epsilon (While e1 e2, m)
        (If e1 (Sequ e2 (While e1 e2)) (Const unit), m).


Definition app_ctx (ctx : context) (s : configuration) :=
  match s with
  | (e, m) => (ctx e, m)
  end.


Inductive red : configuration -> configuration -> Prop :=
  | tete : forall s1 s2 : configuration, red_epsilon s1 s2 -> red s1 s2
  | cont :
      forall (s1 s2 : configuration) (ctx : context),
      red s1 s2 -> is_context ctx -> red (app_ctx ctx s1) (app_ctx ctx s2).
